/* $Id: e2_textiter.c 3 2006-11-01 04:36:04Z tpgww $
 *
 * Copyright (C) 2000, 2002 Paolo Maggi
 * Copyright (C) 2002, 2003 Jeroen Zwartepoorte
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with emelFM2. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/*
 * Parts of this file are from the gedit or glimmer projects.
 */

#include "emelfm2.h"
#include <string.h>
#include "e2_textiter.h"

#define GTK_TEXT_UNKNOWN_CHAR 0xFFFC

/**
 * @brief move foward @a count characters, ignoring any that match supplied flags
 *
 * @param iter text iter referring to current position
 * @param count the no. of chars to move
 * @param skip_invisible TRUE to skip invisible chars
 * @param skip_nontext TRUE to skip non-text chars
 * @param skip_decomp
 * @return
 **/
static void _e2_textiter_forward_chars_with_skipping (
		GtkTextIter *iter,
		gint         count,
		gboolean     skip_invisible,
		gboolean     skip_nontext,
		gboolean     skip_decomp)
{
	gint i;

	g_return_if_fail (count >= 0);

	i = count;

	while (i > 0)
	{
		gboolean ignored = FALSE;

		if (skip_nontext && gtk_text_iter_get_char (iter) == GTK_TEXT_UNKNOWN_CHAR)
			ignored = TRUE;

		if (!ignored && skip_invisible
		    /*&& _gtk_text_btree_char_is_invisible (iter)*/)
			ignored = TRUE;

		if (!ignored && skip_decomp)
		{
			/* being utf8-correct sucks; this accounts for extra
			   offsets coming from canonical decompositions of
			   utf8 characters (e.g. accented characters) which
			   g_utf8_normalize() performs */
			gunichar *decomp;
			gsize decomp_len;
			decomp = g_unicode_canonical_decomposition (
				gtk_text_iter_get_char (iter), &decomp_len);
			i -= (decomp_len - 1);
			g_free (decomp);
		}

		gtk_text_iter_forward_char (iter);

		if (!ignored)
			--i;
	}
}

/**
 * @brief
 *
 * @param start
 * @param lines array
 * @param visible_only TRUE to ignore hidden chars
 * @param slice TRUE to include "unknown" (0xFFFC) chars
 * @param match_start textiter or NULL to try to match the entire line
 * @param match_end textiter
 * @return TRUE if a match was found
 **/
static gboolean _e2_textiter_forward_lines_match (
		const GtkTextIter *start,
		const gchar      **lines,
		gboolean           visible_only,
		gboolean           slice,
		GtkTextIter       *match_start,
		GtkTextIter       *match_end)
{
	GtkTextIter next;
	gchar *line_text;
	const gchar *found;
	gint offset;

	if (*lines == NULL || **lines == '\0')
	{
		if (match_start)
			*match_start = *start;
		if (match_end)
			*match_end = *start;
		return TRUE;
	}

	next = *start;
	gtk_text_iter_forward_line (&next);

	/* No more text in buffer, but *lines is nonempty */
	if (gtk_text_iter_equal (start, &next))
		return FALSE;

	if (slice)
	{
		if (visible_only)
			line_text = gtk_text_iter_get_visible_slice (start, &next);
		else
			line_text = gtk_text_iter_get_slice (start, &next);
	}
	else
	{
		if (visible_only)
			line_text = gtk_text_iter_get_visible_text (start, &next);
		else
			line_text = gtk_text_iter_get_text (start, &next);
	}

	if (match_start) /* if this is the first line we're matching */
	{
		found = e2_utf8_strcasestr (line_text, *lines);
	}
	else
	{
		/* If it's not the first line, we have to match from the
		 * start of the line */
		if (e2_utf8_caseless_match (line_text, *lines, -1, -1))
			found = line_text;
		else
			found = NULL;
	}

	if (found == NULL)
	{
		g_free (line_text);
		return FALSE;
	}

	/* Get offset to start of search string */
	offset = g_utf8_strlen (line_text, found - line_text);

	next = *start;

	/* If match start needs to be returned, set it to the
	 * start of the search string */
	_e2_textiter_forward_chars_with_skipping (&next, offset, visible_only,
		!slice, FALSE);
	if (match_start)
	{
		*match_start = next;
	}

	/* Go to end of search string */
	_e2_textiter_forward_chars_with_skipping (&next, g_utf8_strlen (*lines, -1),
		visible_only, !slice, TRUE);

	g_free (line_text);

	++lines;

	if (match_end)
		*match_end = next;

	/* pass NULL for match_start, since we don't need to find the
	 * start again */
	return _e2_textiter_forward_lines_match (&next, lines, visible_only, slice,
		NULL, match_end);
}

/**
 * @brief
 *
 * @param start
 * @param lines array
 * @param visible_only TRUE to ignore hidden chars
 * @param slice TRUE to include "unknown" (0xFFFC) chars
 * @param match_start textiter or NULL to try to match the entire line
 * @param match_end textiter
 * @return TRUE if a match was found
 **/
static gboolean _e2_textiter_backward_lines_match (
		const GtkTextIter *start,
		const gchar      **lines,
		gboolean           visible_only,
		gboolean           slice,
		GtkTextIter       *match_start,
		GtkTextIter       *match_end)
{
	GtkTextIter line, next;
	gchar *line_text;
	const gchar *found;
	gint offset;

	if (*lines == NULL || **lines == '\0')
	{
		if (match_start)
			*match_start = *start;
		if (match_end)
			*match_end = *start;
		return TRUE;
	}

	line = next = *start;
	if (gtk_text_iter_get_line_offset (&next) == 0)
	{
		if (!gtk_text_iter_backward_line (&next))
			return FALSE;
	}
	else
		gtk_text_iter_set_line_offset (&next, 0);

	if (slice)
	{
		if (visible_only)
			line_text = gtk_text_iter_get_visible_slice (&next, &line);
		else
			line_text = gtk_text_iter_get_slice (&next, &line);
	}
	else
	{
		if (visible_only)
			line_text = gtk_text_iter_get_visible_text (&next, &line);
		else
			line_text = gtk_text_iter_get_text (&next, &line);
	}

	if (match_start) /* if this is the first line we're matching */
	{
		found = e2_utf8_strrcasestr (line_text, *lines);
	}
	else
	{
		/* If it's not the first line, we have to match from the
		 * start of the line */
		if (e2_utf8_caseless_match (line_text, *lines, -1, -1))
			found = line_text;
		else
			found = NULL;
	}

	if (found == NULL)
	{
		g_free (line_text);
		return FALSE;
	}

	/* Get offset to start of search string */
	offset = g_utf8_strlen (line_text, found - line_text);

	_e2_textiter_forward_chars_with_skipping (&next, offset, visible_only,
		!slice, FALSE);

	/* If match start needs to be returned, set it to the
	 * start of the search string */
	if (match_start)
	{
		*match_start = next;
	}

	/* Go to end of search string */
	_e2_textiter_forward_chars_with_skipping (&next, g_utf8_strlen (*lines, -1),
		visible_only, !slice, TRUE);

	g_free (line_text);

	++lines;

	if (match_end)
		*match_end = next;

	/* try to match the rest of the lines forward, passing NULL for
	 * match_start so lines_match will try to match the entire line */
	return _e2_textiter_forward_lines_match (&next, lines, visible_only, slice,
		NULL, match_end);
}

/**
 * @brief Search forward for @a str
 *
 * Same as gtk_text_iter_forward_search(), but supports
 * case-insensitive searching.
 *
 * Any match is returned by setting @a match_start to the first
 * character of the match and @a match_end to the first character
 * after the match. The search will not continue past @a limit.
 * Note that a search is a linear or O(n) operation, so you
 * may wish to use @a limit to avoid locking up your UI on large
 * buffers.
 *
 * If the GTK_SOURCE_SEARCH_VISIBLE_ONLY flag is present, the match
 * may have invisible text interspersed in @a str. i.e. @a str will be a
 * possibly-noncontiguous subset of the matched range. Similarly, if
 * you specify GTK_SOURCE_SEARCH_TEXT_ONLY, the match may have
 * pixbufs or child widgets inside the matched range. If these flags
 * are not given, the match must be exact; the special 0xFFFC
 * character in @a str will match embedded pixbufs or child widgets.
 * If you specify the GTK_SOURCE_SEARCH_CASE_INSENSITIVE flag,
 * the text will be matched regardless of its case.
 *
 * @param  iter GtkTextIter where the search begins
 * @param  str  search string
 * @param  flags  flags affecting how the search is done
 * @param  match_start  return location for start of match, or NULL
 * @param  match_end  return location for end of match, or NULL
 * @param  limit  bound for the search, or NULL for the end of the buffer
 * @return  TRUE if a match was found
 **/
gboolean e2_iter_forward_search (
		const GtkTextIter   *iter,
		const gchar         *str,
		GtkSourceSearchFlags flags,
		GtkTextIter         *match_start,
		GtkTextIter         *match_end,
		const GtkTextIter   *limit)
{
	gchar **lines = NULL;
	GtkTextIter match;
	gboolean retval = FALSE;
	GtkTextIter search;
	gboolean visible_only;
	gboolean slice;

	g_return_val_if_fail (iter != NULL, FALSE);
	g_return_val_if_fail (str != NULL, FALSE);

	if ((flags & GTK_SOURCE_SEARCH_CASE_INSENSITIVE) == 0)
		return gtk_text_iter_forward_search (iter, str, flags,
						     match_start, match_end,
						     limit);

	if (limit && gtk_text_iter_compare (iter, limit) >= 0)
		return FALSE;

	if (*str == '\0')
	{
		/* If we can move one char, return the empty string there */
		match = *iter;

		if (gtk_text_iter_forward_char (&match))
		{
			if (limit && gtk_text_iter_equal (&match, limit))
				return FALSE;

			if (match_start)
				*match_start = match;
			if (match_end)
				*match_end = match;
			return TRUE;
		}
		else
		{
			return FALSE;
		}
	}

	visible_only = (flags & GTK_SOURCE_SEARCH_VISIBLE_ONLY) != 0;
	slice = (flags & GTK_SOURCE_SEARCH_TEXT_ONLY) == 0;

	/* locate all lines */
	lines = e2_utils_str_breakup (str, "\n", -1);
	if (lines == NULL)
		return FALSE;	//FIXME warn user about error

	search = *iter;

	do
	{
		/* This loop has an inefficient worst-case, where
		 * gtk_text_iter_get_text () is called repeatedly on
		 * a single line.
		 */
		GtkTextIter end;

		if (limit && gtk_text_iter_compare (&search, limit) >= 0)
			break;

		if (_e2_textiter_forward_lines_match (&search, (const gchar**)lines,
				 visible_only, slice, &match, &end))
		{
			if (limit == NULL || (//limit &&
					      gtk_text_iter_compare (&end, limit) < 0))
			{
				retval = TRUE;

				if (match_start)
					*match_start = match;
				if (match_end)
					*match_end = end;
			}
			break;
		}
	} while (gtk_text_iter_forward_line (&search));

	g_strfreev (lines);

	return retval;
}

/**
 * @brief Search backward for @a str
 *
 * Same as gtk_text_iter_forward_search(), but supports
 * case-insensitive searching.
 *
 * @param  iter a GtkTextIter where the search begins
 * @param  str search string
 * @param  flags bitmask of flags affecting the search
 * @param  match_start return location for start of match, or NULL
 * @param  match_end  return location for end of match, or NULL
 * @param  limit  location of last possible match_start, or NULL for start of buffer
 *
 * @return TRUE if a match was found
 **/
gboolean e2_iter_backward_search (
		const GtkTextIter   *iter,
		const gchar         *str,
		GtkSourceSearchFlags flags,
		GtkTextIter         *match_start,
		GtkTextIter         *match_end,
		const GtkTextIter   *limit)
{
	gchar **lines = NULL;
	GtkTextIter match;
	gboolean retval = FALSE;
	GtkTextIter search;
	gboolean visible_only;
	gboolean slice;

	g_return_val_if_fail (iter != NULL, FALSE);
	g_return_val_if_fail (str != NULL, FALSE);

	if ((flags & GTK_SOURCE_SEARCH_CASE_INSENSITIVE) == 0)
		return gtk_text_iter_backward_search (iter, str, flags,
						      match_start, match_end, limit);

	if (limit && gtk_text_iter_compare (iter, limit) <= 0)
		return FALSE;

	if (*str == '\0')
	{
		/* If we can move one char, return the empty string there */
		match = *iter;

		if (gtk_text_iter_backward_char (&match))
		{
			if (limit && gtk_text_iter_equal (&match, limit))
				return FALSE;

			if (match_start)
				*match_start = match;
			if (match_end)
				*match_end = match;
			return TRUE;
		}
		else
		{
			return FALSE;
		}
	}

	visible_only = (flags & GTK_SOURCE_SEARCH_VISIBLE_ONLY) != 0;
	slice = (flags & GTK_SOURCE_SEARCH_TEXT_ONLY) == 0;

	/* locate all lines */
	lines = e2_utils_str_breakup (str, "\n", -1);
	if (lines == NULL)
		return FALSE;	//FIXME warn user about error

	search = *iter;

	while (TRUE)
	{
		/* This loop has an inefficient worst-case, where
		 * gtk_text_iter_get_text () is called repeatedly on
		 * a single line */
		GtkTextIter end;

		if (limit && gtk_text_iter_compare (&search, limit) <= 0)
			break;

		if (_e2_textiter_backward_lines_match (&search, (const gchar**)lines,
					  visible_only, slice, &match, &end))
		{
			if (limit == NULL || (//limit &&
				gtk_text_iter_compare (&end, limit) > 0))
			{
				retval = TRUE;

				if (match_start)
					*match_start = match;
				if (match_end)
					*match_end = end;
			}
			break;
		}

		if (gtk_text_iter_get_line_offset (&search) == 0)
		{
			if (!gtk_text_iter_backward_line (&search))
				break;
		}
		else
		{
			gtk_text_iter_set_line_offset (&search, 0);
		}
	}

	g_strfreev (lines);

	return retval;
}
